@@ -60,7 +60,7 @@ class SegmentGroupsAPI(APIClient):
             ...     print(group.as_dict())
 
             Fetch segment groups with query_params filters
-            >>> group_list, _, err = client.zpa.pra_portal.list_portals(
+            >>> group_list, _, err = client.zpa.segment_groups.list_groups(
             ... query_params={'search': 'Group01', 'page': '1', 'page_size': '100'})
             ... if err:
             ...     print(f"Error listing segment groups: {err}")
